#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from __future__ import absolute_import, print_function
import readline
import traceback
from .debuggercli import DebuggerCLI


class ReadlineCompleteHook():
    def __init__(self, completer):
        self._completer = completer
        self._last_contents = (None, None)
        self._last_completions = []

    def complete(self, text, state):
        # pylint: disable=broad-except
        try:
            entire_text = readline.get_line_buffer()[:readline.get_endidx()]
            completions = self._get_completions(entire_text, text)
            return completions[state]
        except Exception:
            traceback.print_exc()
            return []

    def _get_completions(self, entire_text, text):
        if self._last_contents == (entire_text, text):
            return self._last_completions
        self._last_completions = self._completer.complete(entire_text, text)
        self._last_completions.append(None)
        self._last_contents = (entire_text, text)
        return self._last_completions


__setup_performed__ = False


def setup_readline():
    # pylint: disable=global-statement
    global __setup_performed__

    if __setup_performed__:
        return

    debuggercli = DebuggerCLI()
    readline.parse_and_bind("tab: complete")
    readline.set_completer(ReadlineCompleteHook(debuggercli.get_root_completer()).complete)
    readline.set_completer_delims(' \t\n\"\'`@><=;|&')
    __setup_performed__ = True



#############################################################################
# Copyright (c) 2022 Balazs Scheidler <bazsi77@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
# pylint: disable=function-redefined

try:
    from _syslogng import LogParser

except ImportError:
    import warnings
    warnings.warn("You have imported the syslogng package outside of syslog-ng, "
                  "thus some of the functionality is not available. "
                  "Defining fake classes for those exported by the underlying syslog-ng code")

    LogParser = object


class LogParser(LogParser):
    def init(self, options):
        """This method is called when the configuration is initialized

        Args:
            options:
                is a dictionary that contains options passed using the
                options() parameter of the python() destination

        Returns:
            bool: True if successful, False otherwise
        """
        return True

    def deinit(self):
        """This method is called when the configuration is deinitialized"""

    def parse(self, msg):
        """This method is called to 'parse' the message

        Args:
            msg: LogMessage
                is a syslog-ng message that allows dict-like access to
                name-value pairs.

        Returns:
            bool: True if parsing was successful, False to drop the message
        """
        raise NotImplementedError



#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from .lexertoken import LexerToken
from .lexer import Lexer


class TabLexer(Lexer):
    def __init__(self, lexer):
        self._lexer = lexer
        self._current_token = None
        self._end_of_tokens = False
        self._replaced_token = None
        self._buffered_tokens = None
        self._buffer_count = 0
        self._is_last_token = False

    def input(self, text):
        self._lexer.input(text)
        self._end_of_tokens = False
        self._buffered_tokens = None
        self._replaced_token = None

    def get_replaced_token(self):
        return self._replaced_token

    def set_drop_last_token(self, value):
        if value:
            self._buffer_count = 1
        else:
            self._buffer_count = 0

    def token(self):
        if self._end_of_tokens:
            return None
        if not self._is_buffer_initialized():
            self._fill_buffer()

        self._shift_and_inject_tab()
        return self._current_token

    def get_position(self):
        # This method is not implemented as it would require to look ahead
        # to the next token we may or may not have, making the code more
        # difficult. We are not using this anyway.
        raise NotImplementedError

    def _shift_and_inject_tab(self):
        self._shift_from_buffer()

        if self._is_last_token or self._is_current_token_partial():
            self._end_of_tokens = True
            self._replaced_token = self._current_token
            self._current_token = self._constuct_tab_token()

    def _is_current_token_partial(self):
        return (self._current_token is not None and
                (hasattr(self._current_token, 'partial') and
                 self._current_token.partial))

    def _is_buffer_initialized(self):
        return self._buffered_tokens is not None

    def _constuct_tab_token(self):
        if self._replaced_token is not None:
            lexpos = self._replaced_token.lexpos
        else:
            lexpos = self._lexer.get_position()
        return LexerToken(type="TAB", lexpos=lexpos)

    def _shift_from_buffer(self):
        self._fetch_token_to_buffer()
        self._is_last_token = len(self._buffered_tokens) <= self._buffer_count
        self._current_token = self._get_token_from_buffer()

    def _fill_buffer(self):
        self._buffered_tokens = []
        for _ in range(self._buffer_count):
            self._fetch_token_to_buffer()

    def _fetch_token_to_buffer(self):
        token = self._lexer.token()
        if token is not None:
            self._buffered_tokens.append(token)

    def _get_token_from_buffer(self):
        try:
            token = self._buffered_tokens[0]
            self._buffered_tokens = self._buffered_tokens[1:]
            return token
        except IndexError:
            return None



#############################################################################
# Copyright (c) 2023 Attila Szakacs
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from __future__ import annotations

from io import SEEK_END
from gzip import GzipFile
from pathlib import Path
from typing import Optional


class CompressableFileBuffer:
    def __init__(
        self,
        path: Path,
        compress: bool = False,
        compresslevel: Optional[int] = None,
    ) -> None:
        self.__finished = False

        self.__path = path.resolve()
        self.__path.parent.mkdir(parents=True, exist_ok=True)
        self.__path.touch()
        self.__buffer = self.__path.open(mode="r+b")
        self.__buffer.seek(0, SEEK_END)

        self.__gzfile: Optional[GzipFile] = None

        if compress:
            assert compresslevel is not None
            self.__gzfile = GzipFile(
                filename="",
                fileobj=self.__buffer,
                mode="wb",
                compresslevel=compresslevel,
            )

    @property
    def path(self) -> Path:
        return self.__path

    @property
    def gzipfile(self) -> Optional[GzipFile]:
        return self.__gzfile

    def write(self, data: bytes) -> int:
        """Raises OSError on failure."""
        assert not self.__finished
        if self.__gzfile:
            return self.__gzfile.write(data)
        return self.__buffer.write(data)

    def flush(self) -> None:
        if self.__gzfile:
            self.__gzfile.flush()
        self.__buffer.flush()

    def is_empty(self) -> int:
        return self.tell() == 0

    def tell(self) -> int:
        return self.__buffer.tell()

    def close(self) -> None:
        if self.__gzfile:
            self.__gzfile.close()
        self.__buffer.close()

    def finish(self) -> None:
        if self.__finished:
            return

        if self.__gzfile:
            self.__gzfile.close()
            self.__gzfile = None
        self.flush()
        self.__finished = True

    def unlink(self) -> None:
        self.__path.unlink()



#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from __future__ import print_function, absolute_import
from .completer import Completer


class ChoiceCompleter(Completer):
    def __init__(self, choices, prefix=None, suffix=None):
        self._choices = choices
        self._prefix = prefix or ''
        self._suffix = suffix if suffix is not None else ' '

    def complete(self, entire_text, word_to_be_completed):
        if entire_text.startswith(self._prefix):
            return self._handle_input_with_prefix(entire_text, word_to_be_completed)
        return self._handle_input_without_prefix(entire_text)

    def _handle_input_without_prefix(self, entire_input):
        if self._prefix.startswith(entire_input) or entire_input == '':
            return [self._prefix]
        return []

    def _handle_input_with_prefix(self, entire_input, word_to_be_completed):
        entire_input, word_to_be_completed = self._chop_prefixes(entire_input, word_to_be_completed)

        return sorted([self._prefix + choice + self._suffix
                       for choice in self._choices
                       if len(word_to_be_completed) == 0 or choice.startswith(word_to_be_completed)])

    def _chop_prefixes(self, entire_input, word_to_be_completed):
        if word_to_be_completed == entire_input:
            word_to_be_completed = word_to_be_completed[len(self._prefix):]
        entire_input = entire_input[len(self._prefix):]
        return entire_input, word_to_be_completed



#############################################################################
# Copyright (c) 2022 Balazs Scheidler <bazsi77@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
# pylint: disable=unused-import

try:
    from _syslogng import LogMessage

except ImportError:
    import warnings
    warnings.warn("You have imported the syslogng package outside of syslog-ng, "
                  "thus some of the functionality is not available. "
                  "Defining fake classes for those exported by the underlying syslog-ng code")

    class LogMessage(dict):

        # syslog-ng internally stores everything as a sequence of bytes that
        # is strongly "preferred" to be an utf8 string.  It takes a bytes
        # and stores it verbatim.  Strings are converted to utf8 (not the
        # local character set!)

        def __init__(self, msg):
            super().__init__()
            self["MESSAGE"] = msg
            self.bookmark = None

        def __getitem__(self, key):
            if isinstance(key, str):
                key = key.encode('utf8')
            return super().__getitem__(key)

        def __setitem__(self, key, value):
            if isinstance(key, str):
                key = key.encode('utf8')
            if isinstance(value, str):
                value = value.encode('utf8')
            super().__setitem__(key, value)

        def __eq__(self, other):
            return False

        def set_bookmark(self, bookmark):
            self.bookmark = bookmark
